# coding: utf-8

"""


    Generated by: https://openapi-generator.tech
"""

from dataclasses import dataclass
import typing_extensions
import urllib3
from urllib3._collections import HTTPHeaderDict

from gentrace import api_client, exceptions
from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401

from gentrace.model.metadata_value_object import MetadataValueObject
from gentrace.model.run_path_section import RunPathSection
from gentrace.model.local_evaluation import LocalEvaluation
from gentrace.model.step_run import StepRun

# body param


class SchemaForRequestBodyApplicationJson(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "testRuns",
        }
        
        class properties:
            pipelineSlug = schemas.StrSchema
            pipelineId = schemas.StrSchema
            
            
            class collectionMethod(
                schemas.EnumBase,
                schemas.StrSchema
            ):
                
                @schemas.classproperty
                def MANUAL(cls):
                    return cls("manual")
                
                @schemas.classproperty
                def RUNNER(cls):
                    return cls("runner")
            
            
            class branch(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'branch':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class commit(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'commit':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class name(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'name':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class metadata(
                schemas.DictBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneFrozenDictMixin
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def additional_properties() -> typing.Type['MetadataValueObject']:
                        return MetadataValueObject
            
                
                def __getitem__(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, None, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: 'MetadataValueObject',
                ) -> 'metadata':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            triggerRemoteEvals = schemas.BoolSchema
            
            
            class testRuns(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.DictSchema
                    ):
                    
                    
                        class MetaOapg:
                            required = {
                                "stepRuns",
                            }
                            
                            class properties:
                                
                                
                                class id(
                                    schemas.UUIDBase,
                                    schemas.StrBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneStrMixin
                                ):
                                
                                
                                    class MetaOapg:
                                        format = 'uuid'
                                
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[None, str, uuid.UUID, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'id':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                        )
                                caseId = schemas.UUIDSchema
                                name = schemas.StrSchema
                                
                                
                                class inputs(
                                    schemas.DictSchema
                                ):
                                
                                
                                    class MetaOapg:
                                        additional_properties = schemas.AnyTypeSchema
                                    
                                    def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        # dict_instance[name] accessor
                                        return super().__getitem__(name)
                                    
                                    def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        return super().get_item_oapg(name)
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[dict, frozendict.frozendict, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                        **kwargs: typing.Union[MetaOapg.additional_properties, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                                    ) -> 'inputs':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                            **kwargs,
                                        )
                                
                                
                                class expectedOutputs(
                                    schemas.DictSchema
                                ):
                                
                                
                                    class MetaOapg:
                                        additional_properties = schemas.AnyTypeSchema
                                    
                                    def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        # dict_instance[name] accessor
                                        return super().__getitem__(name)
                                    
                                    def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        return super().get_item_oapg(name)
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[dict, frozendict.frozendict, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                        **kwargs: typing.Union[MetaOapg.additional_properties, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                                    ) -> 'expectedOutputs':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                            **kwargs,
                                        )
                                
                                
                                class metadata(
                                    schemas.DictBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneFrozenDictMixin
                                ):
                                
                                
                                    class MetaOapg:
                                        
                                        @staticmethod
                                        def additional_properties() -> typing.Type['MetadataValueObject']:
                                            return MetadataValueObject
                                
                                    
                                    def __getitem__(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                                        # dict_instance[name] accessor
                                        return super().__getitem__(name)
                                    
                                    def get_item_oapg(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                                        return super().get_item_oapg(name)
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[dict, frozendict.frozendict, None, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                        **kwargs: 'MetadataValueObject',
                                    ) -> 'metadata':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                            **kwargs,
                                        )
                                
                                
                                class stepRuns(
                                    schemas.ListSchema
                                ):
                                
                                
                                    class MetaOapg:
                                        
                                        @staticmethod
                                        def items() -> typing.Type['StepRun']:
                                            return StepRun
                                
                                    def __new__(
                                        cls,
                                        _arg: typing.Union[typing.Tuple['StepRun'], typing.List['StepRun']],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'stepRuns':
                                        return super().__new__(
                                            cls,
                                            _arg,
                                            _configuration=_configuration,
                                        )
                                
                                    def __getitem__(self, i: int) -> 'StepRun':
                                        return super().__getitem__(i)
                                
                                
                                class evaluations(
                                    schemas.ListSchema
                                ):
                                
                                
                                    class MetaOapg:
                                        
                                        @staticmethod
                                        def items() -> typing.Type['LocalEvaluation']:
                                            return LocalEvaluation
                                
                                    def __new__(
                                        cls,
                                        _arg: typing.Union[typing.Tuple['LocalEvaluation'], typing.List['LocalEvaluation']],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'evaluations':
                                        return super().__new__(
                                            cls,
                                            _arg,
                                            _configuration=_configuration,
                                        )
                                
                                    def __getitem__(self, i: int) -> 'LocalEvaluation':
                                        return super().__getitem__(i)
                                
                                
                                class path(
                                    schemas.ListBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneTupleMixin
                                ):
                                
                                
                                    class MetaOapg:
                                        
                                        @staticmethod
                                        def items() -> typing.Type['RunPathSection']:
                                            return RunPathSection
                                
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[list, tuple, None, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'path':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                        )
                                
                                
                                class error(
                                    schemas.StrBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneStrMixin
                                ):
                                
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[None, str, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'error':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                        )
                                __annotations__ = {
                                    "id": id,
                                    "caseId": caseId,
                                    "name": name,
                                    "inputs": inputs,
                                    "expectedOutputs": expectedOutputs,
                                    "metadata": metadata,
                                    "stepRuns": stepRuns,
                                    "evaluations": evaluations,
                                    "path": path,
                                    "error": error,
                                }
                        
                        stepRuns: MetaOapg.properties.stepRuns
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["caseId"]) -> MetaOapg.properties.caseId: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["inputs"]) -> MetaOapg.properties.inputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["expectedOutputs"]) -> MetaOapg.properties.expectedOutputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["metadata"]) -> MetaOapg.properties.metadata: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["evaluations"]) -> MetaOapg.properties.evaluations: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["path"]) -> MetaOapg.properties.path: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["error"]) -> MetaOapg.properties.error: ...
                        
                        @typing.overload
                        def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
                        
                        def __getitem__(self, name: typing.Union[typing_extensions.Literal["id", "caseId", "name", "inputs", "expectedOutputs", "metadata", "stepRuns", "evaluations", "path", "error", ], str]):
                            # dict_instance[name] accessor
                            return super().__getitem__(name)
                        
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> typing.Union[MetaOapg.properties.id, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["caseId"]) -> typing.Union[MetaOapg.properties.caseId, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> typing.Union[MetaOapg.properties.name, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["inputs"]) -> typing.Union[MetaOapg.properties.inputs, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["expectedOutputs"]) -> typing.Union[MetaOapg.properties.expectedOutputs, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["metadata"]) -> typing.Union[MetaOapg.properties.metadata, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["evaluations"]) -> typing.Union[MetaOapg.properties.evaluations, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["path"]) -> typing.Union[MetaOapg.properties.path, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["error"]) -> typing.Union[MetaOapg.properties.error, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
                        
                        def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["id", "caseId", "name", "inputs", "expectedOutputs", "metadata", "stepRuns", "evaluations", "path", "error", ], str]):
                            return super().get_item_oapg(name)
                        
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, ],
                            stepRuns: typing.Union[MetaOapg.properties.stepRuns, list, tuple, ],
                            id: typing.Union[MetaOapg.properties.id, None, str, uuid.UUID, schemas.Unset] = schemas.unset,
                            caseId: typing.Union[MetaOapg.properties.caseId, str, uuid.UUID, schemas.Unset] = schemas.unset,
                            name: typing.Union[MetaOapg.properties.name, str, schemas.Unset] = schemas.unset,
                            inputs: typing.Union[MetaOapg.properties.inputs, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
                            expectedOutputs: typing.Union[MetaOapg.properties.expectedOutputs, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
                            metadata: typing.Union[MetaOapg.properties.metadata, dict, frozendict.frozendict, None, schemas.Unset] = schemas.unset,
                            evaluations: typing.Union[MetaOapg.properties.evaluations, list, tuple, schemas.Unset] = schemas.unset,
                            path: typing.Union[MetaOapg.properties.path, list, tuple, None, schemas.Unset] = schemas.unset,
                            error: typing.Union[MetaOapg.properties.error, None, str, schemas.Unset] = schemas.unset,
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                stepRuns=stepRuns,
                                id=id,
                                caseId=caseId,
                                name=name,
                                inputs=inputs,
                                expectedOutputs=expectedOutputs,
                                metadata=metadata,
                                evaluations=evaluations,
                                path=path,
                                error=error,
                                _configuration=_configuration,
                                **kwargs,
                            )
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'testRuns':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            __annotations__ = {
                "pipelineSlug": pipelineSlug,
                "pipelineId": pipelineId,
                "collectionMethod": collectionMethod,
                "branch": branch,
                "commit": commit,
                "name": name,
                "metadata": metadata,
                "triggerRemoteEvals": triggerRemoteEvals,
                "testRuns": testRuns,
            }
    
    testRuns: MetaOapg.properties.testRuns
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["pipelineSlug"]) -> MetaOapg.properties.pipelineSlug: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["pipelineId"]) -> MetaOapg.properties.pipelineId: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["collectionMethod"]) -> MetaOapg.properties.collectionMethod: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["branch"]) -> MetaOapg.properties.branch: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["commit"]) -> MetaOapg.properties.commit: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["metadata"]) -> MetaOapg.properties.metadata: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["triggerRemoteEvals"]) -> MetaOapg.properties.triggerRemoteEvals: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["testRuns"]) -> MetaOapg.properties.testRuns: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["pipelineSlug", "pipelineId", "collectionMethod", "branch", "commit", "name", "metadata", "triggerRemoteEvals", "testRuns", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["pipelineSlug"]) -> typing.Union[MetaOapg.properties.pipelineSlug, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["pipelineId"]) -> typing.Union[MetaOapg.properties.pipelineId, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["collectionMethod"]) -> typing.Union[MetaOapg.properties.collectionMethod, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["branch"]) -> typing.Union[MetaOapg.properties.branch, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["commit"]) -> typing.Union[MetaOapg.properties.commit, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> typing.Union[MetaOapg.properties.name, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["metadata"]) -> typing.Union[MetaOapg.properties.metadata, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["triggerRemoteEvals"]) -> typing.Union[MetaOapg.properties.triggerRemoteEvals, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["testRuns"]) -> MetaOapg.properties.testRuns: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["pipelineSlug", "pipelineId", "collectionMethod", "branch", "commit", "name", "metadata", "triggerRemoteEvals", "testRuns", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        testRuns: typing.Union[MetaOapg.properties.testRuns, list, tuple, ],
        pipelineSlug: typing.Union[MetaOapg.properties.pipelineSlug, str, schemas.Unset] = schemas.unset,
        pipelineId: typing.Union[MetaOapg.properties.pipelineId, str, schemas.Unset] = schemas.unset,
        collectionMethod: typing.Union[MetaOapg.properties.collectionMethod, str, schemas.Unset] = schemas.unset,
        branch: typing.Union[MetaOapg.properties.branch, None, str, schemas.Unset] = schemas.unset,
        commit: typing.Union[MetaOapg.properties.commit, None, str, schemas.Unset] = schemas.unset,
        name: typing.Union[MetaOapg.properties.name, None, str, schemas.Unset] = schemas.unset,
        metadata: typing.Union[MetaOapg.properties.metadata, dict, frozendict.frozendict, None, schemas.Unset] = schemas.unset,
        triggerRemoteEvals: typing.Union[MetaOapg.properties.triggerRemoteEvals, bool, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaForRequestBodyApplicationJson':
        return super().__new__(
            cls,
            *_args,
            testRuns=testRuns,
            pipelineSlug=pipelineSlug,
            pipelineId=pipelineId,
            collectionMethod=collectionMethod,
            branch=branch,
            commit=commit,
            name=name,
            metadata=metadata,
            triggerRemoteEvals=triggerRemoteEvals,
            _configuration=_configuration,
            **kwargs,
        )


request_body_any_type = api_client.RequestBody(
    content={
        'application/json': api_client.MediaType(
            schema=SchemaForRequestBodyApplicationJson),
    },
    required=True,
)


class SchemaFor200ResponseBodyApplicationJson(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "resultId",
        }
        
        class properties:
            resultId = schemas.UUIDSchema
            __annotations__ = {
                "resultId": resultId,
            }
    
    resultId: MetaOapg.properties.resultId
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["resultId"]) -> MetaOapg.properties.resultId: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["resultId", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["resultId"]) -> MetaOapg.properties.resultId: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["resultId", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        resultId: typing.Union[MetaOapg.properties.resultId, str, uuid.UUID, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJson':
        return super().__new__(
            cls,
            *_args,
            resultId=resultId,
            _configuration=_configuration,
            **kwargs,
        )


class SchemaFor200ResponseBodyApplicationJsonCharsetutf8(
    schemas.DictSchema
):


    class MetaOapg:
        required = {
            "resultId",
        }
        
        class properties:
            resultId = schemas.UUIDSchema
            __annotations__ = {
                "resultId": resultId,
            }
    
    resultId: MetaOapg.properties.resultId
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["resultId"]) -> MetaOapg.properties.resultId: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["resultId", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["resultId"]) -> MetaOapg.properties.resultId: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["resultId", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        resultId: typing.Union[MetaOapg.properties.resultId, str, uuid.UUID, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJsonCharsetutf8':
        return super().__new__(
            cls,
            *_args,
            resultId=resultId,
            _configuration=_configuration,
            **kwargs,
        )


@dataclass
class ApiResponseFor200(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor200ResponseBodyApplicationJson,
        SchemaFor200ResponseBodyApplicationJsonCharsetutf8,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_200 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor200,
    content={
        'application/json': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJson),
        'application/json; charset=utf-8': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJsonCharsetutf8),
    },
)
_all_accept_content_types = (
    'application/json',
    'application/json; charset=utf-8',
)


class BaseApi(api_client.Api):
    @typing.overload
    def _v1_test_result_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def _v1_test_result_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def _v1_test_result_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def _v1_test_result_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def _v1_test_result_post_oapg(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        """
        Create a new test result from runs
        :param skip_deserialization: If true then api_response.response will be set but
            api_response.body and api_response.headers will not be deserialized into schema
            class instances
        """
        used_path = path.value

        _headers = HTTPHeaderDict()
        # TODO add cookie handling
        if accept_content_types:
            for accept_content_type in accept_content_types:
                _headers.add('Accept', accept_content_type)

        if body is schemas.unset:
            raise exceptions.ApiValueError(
                'The required body parameter has an invalid value of: unset. Set a valid value instead')
        _fields = None
        _body = None
        serialized_data = request_body_any_type.serialize(body, content_type)
        _headers.add('Content-Type', content_type)
        if 'fields' in serialized_data:
            _fields = serialized_data['fields']
        elif 'body' in serialized_data:
            _body = serialized_data['body']
        response = self.api_client.call_api(
            resource_path=used_path,
            method='post'.upper(),
            headers=_headers,
            fields=_fields,
            body=_body,
            auth_settings=_auth,
            stream=stream,
            timeout=timeout,
        )

        if skip_deserialization:
            api_response = api_client.ApiResponseWithoutDeserialization(response=response)
        else:
            response_for_status = _status_code_to_response.get(str(response.status))
            if response_for_status:
                api_response = response_for_status.deserialize(response, self.api_client.configuration)
            else:
                api_response = api_client.ApiResponseWithoutDeserialization(response=response)

        if not 200 <= response.status <= 299:
            raise exceptions.ApiException(
                status=response.status,
                reason=response.reason,
                api_response=api_response
            )

        return api_response


class V1TestResultPost(BaseApi):
    # this class is used by api classes that refer to endpoints with operationId fn names

    @typing.overload
    def v1_test_result_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def v1_test_result_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def v1_test_result_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def v1_test_result_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def v1_test_result_post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v1_test_result_post_oapg(
            body=body,
            content_type=content_type,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )


class ApiForpost(BaseApi):
    # this class is used by api classes that refer to endpoints by path and http method names

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: typing_extensions.Literal["application/json"] = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...


    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        skip_deserialization: typing_extensions.Literal[True],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = ...,
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def post(
        self,
        body: typing.Union[SchemaForRequestBodyApplicationJson,dict, frozendict.frozendict, ],
        content_type: str = 'application/json',
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v1_test_result_post_oapg(
            body=body,
            content_type=content_type,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )





import asyncio
import json
import logging
import os
import socket
import threading
import uuid
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, Tuple
from urllib.parse import urlparse

import websockets
from gentrace.api_client import ApiClient
from gentrace.apis.tags.v1_api import V1Api
from gentrace.configuration import Configuration
from gentrace.providers.init import (
    GENTRACE_CONFIG_STATE,
)
from gentrace.providers.pipeline import Pipeline
from gentrace.providers.utils import log_debug, log_info, log_warn
import pystache
from gentrace.providers.evaluation import update_test_result_with_runners
from websockets.client import connect as ws_connect
from asyncio import Semaphore
import requests

T = TypeVar("T")

# Context management for overrides
overrides_context: ContextVar[Dict[str, Any]] = ContextVar("overrides", default={})

@contextmanager
def override_context(overrides: Dict[str, Any]):
    """Context manager for handling overrides during test execution."""
    token = overrides_context.set(overrides)
    try:
        yield
    finally:
        overrides_context.reset(token)

def get_value(name: str, default_value: T) -> T:
    """Get a value from the current context or return default"""
    overrides = overrides_context.get()
    value = overrides.get(name, default_value)
    return value

# Global state for interactions and test suites
interactions: Dict[str, Dict] = {}
test_suites: Dict[str, Dict] = {}
listeners: Dict[str, Callable] = {}

def define_interaction(interaction: Dict) -> Callable:
    """Define a new interaction for testing."""
    # Convert parameters to dict format for API compatibility
    api_parameters = []
    if "parameters" in interaction:
        for param in interaction["parameters"]:
            param_dict = {
                "name": param.name,
                "type": param.type,
                "defaultValue": param.default_value,
            }
            if isinstance(param, TemplateParameter):
                param_dict["variables"] = param.variables
            elif isinstance(param, EnumParameter):
                param_dict["options"] = param.options
            api_parameters.append(param_dict)
    
    # Create API-compatible interaction object
    api_interaction = {
        "name": interaction["name"],
        "fn": interaction["fn"],
        "parameters": api_parameters,
    }
    if "inputType" in interaction:
        api_interaction["inputType"] = interaction["inputType"]
    
    interactions[interaction["name"]] = api_interaction
    for listener in listeners.values():
        listener({
            "type": "register-interaction",
            "interaction": api_interaction
        })
    return interaction["fn"]

def define_test_suite(test_suite: Dict) -> Callable:
    """Define a new test suite."""
    test_suites[test_suite["name"]] = test_suite
    for listener in listeners.values():
        listener({
            "type": "register-test-suite",
            "test_suite": test_suite
        })
    return test_suite["fn"]

class BaseParameter:
    def __init__(self, name: str, default_value: Any):
        self.name = name
        self.default_value = default_value
        self.type = "base"

    def get_value(self) -> Any:
        return get_value(self.name, self.default_value)

class TemplateParameter(BaseParameter):
    def __init__(self, name: str, default_value: str, variables: Optional[List[Dict]] = None):
        super().__init__(name, default_value)
        self.type = "template"
        self.variables = variables or []

    def render(self, values: Dict[str, Any]) -> str:
        template = self.get_value()
        return pystache.render(template, values)

class NumericParameter(BaseParameter):
    def __init__(self, name: str, default_value: Union[int, float]):
        super().__init__(name, default_value)
        self.type = "numeric"

class EnumParameter(BaseParameter):
    def __init__(self, name: str, default_value: str, options: List[str]):
        super().__init__(name, default_value)
        self.type = "enum"
        self.options = options

class StringParameter(BaseParameter):
    def __init__(self, name: str, default_value: str):
        super().__init__(name, default_value)
        self.type = "string"

def template_parameter(config: Dict[str, Any]) -> TemplateParameter:
    """Create a template parameter."""
    return TemplateParameter(
        name=config["name"],
        default_value=config["defaultValue"],
        variables=config.get("variables", [])
    )

def numeric_parameter(config: Dict[str, Any]) -> NumericParameter:
    """Create a numeric parameter."""
    return NumericParameter(
        name=config["name"],
        default_value=config["defaultValue"]
    )

def enum_parameter(config: Dict[str, Any]) -> EnumParameter:
    """Create an enum parameter."""
    return EnumParameter(
        name=config["name"],
        default_value=config["defaultValue"],
        options=config["options"]
    )

def string_parameter(config: Dict[str, Any]) -> StringParameter:
    """Create a string parameter."""
    return StringParameter(
        name=config["name"],
        default_value=config["defaultValue"]
    )

def get_ws_base_path() -> str:
    """Get the WebSocket base path based on configuration."""
    if not GENTRACE_CONFIG_STATE["global_gentrace_config"]:
        return "wss://gentrace.ai/ws"
    
    parsed = urlparse(GENTRACE_CONFIG_STATE["global_gentrace_config"].host)
    if "localhost" in parsed.netloc:
        return "ws://localhost:3001"
    
    base_path = f"wss://{parsed.netloc}/ws"
    return base_path

async def handle_message(message: Dict, transport: Dict) -> None:
    """Handle incoming WebSocket messages."""
    message_type = message.get("type")
    print(f"Received message type: {message_type}")
    
    if message_type == "environment-details":
        print("Processing environment details request")
        # Convert parameters to API format for each interaction
        api_interactions = []
        for interaction in interactions.values():
            api_interaction = {
                "name": interaction["name"],
                "hasValidation": bool(interaction.get("inputType")),
                "parameters": interaction.get("parameters", [])
            }
            api_interactions.append(api_interaction)
            
        await send_message({
            "type": "environment-details",
            "interactions": api_interactions,
            "testSuites": [
                {"name": suite["name"]}
                for suite in test_suites.values()
            ]
        }, transport)
    
    elif message_type == "run-interaction-input-validation":
        print(f"Running input validation for interaction: {message['interactionName']}")
        interaction = interactions.get(message["interactionName"])
        if not interaction:
            await send_message({
                "type": "run-interaction-input-validation-results",
                "id": message["id"],
                "interactionName": message["interactionName"],
                "data": [
                    {
                        "id": tc["id"],
                        "status": "failure",
                        "error": f"Interaction {message['interactionName']} not found"
                    }
                    for tc in message["data"]
                ]
            }, transport)
            return

        results = []
        for test_case in message["data"]:
            try:
                if interaction.get("inputType"):
                    interaction["inputType"].parse_obj(test_case["inputs"])
                    results.append({"id": test_case["id"], "status": "success"})
                else:
                    results.append({
                        "id": test_case["id"],
                        "status": "failure",
                        "error": "No input validator found"
                    })
            except Exception as e:
                results.append({
                    "id": test_case["id"],
                    "status": "failure",
                    "error": str(e)
                })
        
        await send_message({
            "type": "run-interaction-input-validation-results",
            "id": message["id"],
            "interactionName": message["interactionName"],
            "data": results
        }, transport)

    elif message_type == "run-test-interaction":
        print(f"Running test interaction: {message['interactionName']} with overrides: {message.get('overrides', {})}")
        
        pipeline = Pipeline({"id": message["pipelineId"]})
        interaction = interactions.get(message["interactionName"])
        
        if not interaction:
            log_warn(f"Interaction {message['interactionName']} not found")
            return
            
        with override_context(message.get("overrides", {})):
            # Get parallelism from message, default to 1 if not specified
            parallelism = message.get("parallelism", 1)
            semaphore = Semaphore(parallelism)
            
            async def run_with_semaphore(test_case: Dict):
                async with semaphore:
                    return await run_test_case_through_interaction(pipeline, interaction, test_case)
            
            # Run all test cases in parallel with semaphore control
            try:
                results = await asyncio.gather(
                    *[run_with_semaphore(test_case) for test_case in message["data"]],
                    return_exceptions=True
                )
                print(f"Completed test cases execution with {len(results)} results")
               
                # Process results and update test results
                error_results = []
                for result in results:
                    if isinstance(result, Exception):
                        log_warn(f"Error in parallel execution: {result}")
                        error_results.append(str(result))
                        continue
                    runner, test_case = result
                    update_test_result_with_runners(message["testJobId"], [(runner, test_case)])
                
                # Send final status update
                if not GENTRACE_CONFIG_STATE["global_gentrace_config"]:
                    raise ValueError("Gentrace configuration not initialized")
                
                config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
                host = config.host or "https://gentrace.ai/api"
                
                # Get auth settings from config
                auth_settings = config.auth_settings()
                
                # Prepare headers from config with auth
                headers = {
                    "Content-Type": "application/json",
                }
                
                # Add bearer token if available
                if 'bearerAuth' in auth_settings:
                    headers[auth_settings['bearerAuth']['key']] = auth_settings['bearerAuth']['value']
                
                # Make status update request
                status_url = f"{host}/v1/test-result/status"
                status_payload = {
                    "id": message["testJobId"],
                    "finished": True
                }
                
                response = requests.post(
                    status_url,
                    headers=headers,
                    json=status_payload
                )
                
                if not response.ok:
                    log_warn(f"Failed to update test status: {response.text}")
            except Exception as e:
                log_warn(f"Error in parallel test execution: {e}")

    elif message_type == "run-test-suite":
        print(f"Running test suite: {message['testSuiteName']}")
        test_suite = test_suites.get(message["testSuiteName"])
        if not test_suite:
            log_warn(f"Test suite {message['testSuiteName']} not found")
            return
            
        try:
            await test_suite["fn"]()
            await update_test_status(message["testJobId"], True)
        except Exception as e:
            await update_test_status(message["testJobId"], True, str(e))

async def send_message(message: Dict, transport: Dict) -> None:
    """Send a message through the WebSocket transport."""
    if transport["type"] == "ws":
        if not transport.get("pluginId"):
            transport["messageQueue"].append(message)
            return
            
        if transport.get("isClosed"):
            return
            
        await transport["ws"].send(json.dumps({
            "id": str(uuid.uuid4()),
            "for": transport["pluginId"],
            "data": message
        }))
    else:
        transport["sendResponse"](message)


async def update_test_status(test_job_id: str, finished: bool, error: Optional[str] = None) -> None:
    """Update test job status in the Gentrace API."""
    if not GENTRACE_CONFIG_STATE["global_gentrace_config"]:
        raise ValueError("Gentrace configuration not initialized")
        
    api_client = ApiClient(configuration=GENTRACE_CONFIG_STATE["global_gentrace_config"])
    api = V1Api(api_client=api_client)
    
    await api.v1_test_result_status_post({
        "id": test_job_id,
        "finished": finished,
        "error": error
    })

async def run_websocket(environment_name: Optional[str] = None) -> None:
    """Run the WebSocket connection to the Gentrace server."""
    if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
        raise ValueError("Gentrace API key not set")
    
    ws_base_path = get_ws_base_path()
    env = environment_name or GENTRACE_CONFIG_STATE.get("GENTRACE_ENVIRONMENT_NAME") or socket.gethostname()
    
    transport = {
        "type": "ws",
        "pluginId": None,
        "messageQueue": [],
        "isClosed": False
    }

    async def setup():
        """Setup function to register existing interactions and test suites."""
        # Register existing interactions
        for interaction in interactions.values():
            parameters = []
            for param in interaction.get("parameters", []):
                param_copy = param.copy()
                # Remove any function properties
                param_copy.pop("render", None)  # For template parameters
                param_copy.pop("getValue", None)  # For numeric/enum parameters
                parameters.append(param_copy)
            
            await send_message({
                "type": "register-interaction",
                "interaction": {
                    "name": interaction["name"],
                    "hasValidation": bool(interaction.get("inputType")),
                    "parameters": parameters
                }
            }, transport)
        
        # Register existing test suites
        for test_suite in test_suites.values():
            await send_message({
                "type": "register-test-suite",
                "testSuite": {
                    "name": test_suite["name"]
                }
            }, transport)

    websocket = await ws_connect(ws_base_path)
    transport["ws"] = websocket
    
    try:
        print("WebSocket connection opened, sending setup message")
        setup_message = {
            "id": str(uuid.uuid4()),
            "init": "test-job-runner", 
            "data": {
                "type": "setup",
                "environmentName": env,
                "apiKey": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]
            }
        }
        await websocket.send(json.dumps(setup_message))

        # Start ping and heartbeat tasks
        async def send_ping():
            while not transport["isClosed"]:
                try:
                    print("sending ping")
                    await websocket.send(json.dumps({
                        "id": str(uuid.uuid4()),
                        "ping": True
                    }))
                    await asyncio.sleep(30)  # 30 seconds interval
                except Exception as e:
                    if not transport["isClosed"]:
                        break

        async def send_heartbeat():
            while not transport["isClosed"]:
                try:
                    print("sending heartbeat")
                    await send_message({
                        "type": "heartbeat"
                    }, transport)
                    await asyncio.sleep(30)  # 30 seconds interval
                except Exception as e:
                    if not transport["isClosed"]:
                        break

        ping_task = asyncio.create_task(send_ping())
        heartbeat_task = asyncio.create_task(send_heartbeat())

        try:
            while True:
                message = await websocket.recv()
                message_data = json.loads(message)
                print("WebSocket message received:", message)
                
                if "pluginId" in message_data.get("data", {}):
                    transport["pluginId"] = message_data["data"]["pluginId"]
                    # Process queued messages and setup registrations
                    for queued_message in transport["messageQueue"]:
                        await send_message(queued_message, transport)
                    transport["messageQueue"].clear()
                    await setup()  # Register existing interactions and test suites
                    continue

                if message_data.get("error"):
                    log_warn(f"WebSocket error: {message_data['error']}")
                    continue
                    
                try:
                    await handle_message(message_data["data"], transport)
                except Exception as e:
                    log_warn(f"Error handling message: {e}")
                    
        except websockets.exceptions.ConnectionClosed:
            print("WebSocket connection closed")
            transport["isClosed"] = True
        except Exception as e:
            print(f"Error in WebSocket message handler: {e}")
            transport["isClosed"] = True
        finally:
            transport["isClosed"] = True
            ping_task.cancel()
            heartbeat_task.cancel()
            try:
                await asyncio.gather(ping_task, heartbeat_task, return_exceptions=True)
            except asyncio.CancelledError:
                pass
    finally:
        await websocket.close()

async def listen_inner(environment_name: Optional[str] = None, retries: int = 0) -> None:
    """Internal function to handle WebSocket connection with retries."""
    if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
        raise ValueError("Gentrace API key not set")
        
    try:
        await run_websocket(environment_name)
    except Exception as e:
        log_warn(f"Error in WebSocket connection: {e}")
        await asyncio.sleep(min(2 ** retries * 0.25, 10))
        await listen_inner(environment_name, retries + 1)

def listen(environment_name: Optional[str] = None) -> None:
    """Start listening for test jobs from the Gentrace server."""
    asyncio.run(listen_inner(environment_name))

async def handle_webhook(body: Dict, send_response: Optional[Callable[[Dict], None]] = None) -> Dict:
    """
    Handle webhook requests for test job execution.
    
    Args:
        body: The webhook request body
        send_response: Optional callback function to send response back to the webhook caller
    
    Returns:
        Dict containing the response data
    """
    print("Gentrace HTTP message received:", body)
    
    response_data = {}
    
    # Create an HTTP transport for webhook communication
    transport = {
        "type": "http",
        "sendResponse": send_response if send_response else lambda x: response_data.update(x)
    }
    
    try:
        # Process the webhook message using the existing message handler
        await handle_message(body, transport)
        
        # If no send_response provided, return the collected response
        if not send_response:
            return response_data
            
    except Exception as e:
        error_response = {
            "type": "error",
            "error": str(e)
        }
        
        # If send_response provided, use it, otherwise return error
        if send_response:
            send_response(error_response)
        return error_response

async def run_test_case_through_interaction(pipeline: Pipeline, interaction: Dict, test_case: Dict) -> Tuple[Any, Dict]:
    """Run a single test case through an interaction and return the runner and test case."""
    print(f"Running test case {test_case['id']}")
    runner = pipeline.start()
    try:
        await runner.ameasure(interaction["fn"], inputs=test_case["inputs"])
        print(f"Successfully completed test case {test_case['id']}")
    except Exception as e:
        print(f"Failed to run test case {test_case['id']}: {str(e)}")
        runner.set_error(str(e))
    return runner, test_case



# coding: utf-8

"""


    Generated by: https://openapi-generator.tech
"""

from dataclasses import dataclass
import typing_extensions
import urllib3
from urllib3._collections import HTTPHeaderDict

from gentrace import api_client, exceptions
from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401

from gentrace.model.dataset_v2 import DatasetV2

from . import path

# Query params
PipelineSlugSchema = schemas.StrSchema
PipelineIdSchema = schemas.UUIDSchema
ArchivedSchema = schemas.BoolSchema
RequestRequiredQueryParams = typing_extensions.TypedDict(
    'RequestRequiredQueryParams',
    {
    }
)
RequestOptionalQueryParams = typing_extensions.TypedDict(
    'RequestOptionalQueryParams',
    {
        'pipelineSlug': typing.Union[PipelineSlugSchema, str, ],
        'pipelineId': typing.Union[PipelineIdSchema, str, uuid.UUID, ],
        'archived': typing.Union[ArchivedSchema, bool, ],
    },
    total=False
)


class RequestQueryParams(RequestRequiredQueryParams, RequestOptionalQueryParams):
    pass


request_query_pipeline_slug = api_client.QueryParameter(
    name="pipelineSlug",
    style=api_client.ParameterStyle.FORM,
    schema=PipelineSlugSchema,
    explode=True,
)
request_query_pipeline_id = api_client.QueryParameter(
    name="pipelineId",
    style=api_client.ParameterStyle.FORM,
    schema=PipelineIdSchema,
    explode=True,
)
request_query_archived = api_client.QueryParameter(
    name="archived",
    style=api_client.ParameterStyle.FORM,
    schema=ArchivedSchema,
    explode=True,
)
_auth = [
    'bearerAuth',
]


class SchemaFor200ResponseBodyApplicationJson(
    schemas.DictSchema
):


    class MetaOapg:
        
        class properties:
            
            
            class data(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def items() -> typing.Type['DatasetV2']:
                        return DatasetV2
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['DatasetV2'], typing.List['DatasetV2']],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'data':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> 'DatasetV2':
                    return super().__getitem__(i)
            __annotations__ = {
                "data": data,
            }
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["data"]) -> MetaOapg.properties.data: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["data"]) -> typing.Union[MetaOapg.properties.data, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        data: typing.Union[MetaOapg.properties.data, list, tuple, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJson':
        return super().__new__(
            cls,
            *_args,
            data=data,
            _configuration=_configuration,
            **kwargs,
        )


class SchemaFor200ResponseBodyApplicationJsonCharsetutf8(
    schemas.DictSchema
):


    class MetaOapg:
        
        class properties:
            
            
            class data(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def items() -> typing.Type['DatasetV2']:
                        return DatasetV2
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['DatasetV2'], typing.List['DatasetV2']],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'data':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> 'DatasetV2':
                    return super().__getitem__(i)
            __annotations__ = {
                "data": data,
            }
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["data"]) -> MetaOapg.properties.data: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["data"]) -> typing.Union[MetaOapg.properties.data, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["data", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        data: typing.Union[MetaOapg.properties.data, list, tuple, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'SchemaFor200ResponseBodyApplicationJsonCharsetutf8':
        return super().__new__(
            cls,
            *_args,
            data=data,
            _configuration=_configuration,
            **kwargs,
        )


@dataclass
class ApiResponseFor200(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: typing.Union[
        SchemaFor200ResponseBodyApplicationJson,
        SchemaFor200ResponseBodyApplicationJsonCharsetutf8,
    ]
    headers: schemas.Unset = schemas.unset


_response_for_200 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor200,
    content={
        'application/json': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJson),
        'application/json; charset=utf-8': api_client.MediaType(
            schema=SchemaFor200ResponseBodyApplicationJsonCharsetutf8),
    },
)


@dataclass
class ApiResponseFor400(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: schemas.Unset = schemas.unset
    headers: schemas.Unset = schemas.unset


_response_for_400 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor400,
)


@dataclass
class ApiResponseFor500(api_client.ApiResponse):
    response: urllib3.HTTPResponse
    body: schemas.Unset = schemas.unset
    headers: schemas.Unset = schemas.unset


_response_for_500 = api_client.OpenApiResponse(
    response_cls=ApiResponseFor500,
)
_status_code_to_response = {
    '200': _response_for_200,
    '400': _response_for_400,
    '500': _response_for_500,
}
_all_accept_content_types = (
    'application/json',
    'application/json; charset=utf-8',
)


class BaseApi(api_client.Api):
    @typing.overload
    def _v2_datasets_get_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def _v2_datasets_get_oapg(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def _v2_datasets_get_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def _v2_datasets_get_oapg(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        """
        List datasets
        :param skip_deserialization: If true then api_response.response will be set but
            api_response.body and api_response.headers will not be deserialized into schema
            class instances
        """
        self._verify_typed_dict_inputs_oapg(RequestQueryParams, query_params)
        used_path = path.value

        prefix_separator_iterator = None
        for parameter in (
            request_query_pipeline_slug,
            request_query_pipeline_id,
            request_query_archived,
        ):
            parameter_data = query_params.get(parameter.name, schemas.unset)
            if parameter_data is schemas.unset:
                continue
            if prefix_separator_iterator is None:
                prefix_separator_iterator = parameter.get_prefix_separator_iterator()
            serialized_data = parameter.serialize(parameter_data, prefix_separator_iterator)
            for serialized_value in serialized_data.values():
                used_path += serialized_value

        _headers = HTTPHeaderDict()
        # TODO add cookie handling
        if accept_content_types:
            for accept_content_type in accept_content_types:
                _headers.add('Accept', accept_content_type)

        response = self.api_client.call_api(
            resource_path=used_path,
            method='get'.upper(),
            headers=_headers,
            auth_settings=_auth,
            stream=stream,
            timeout=timeout,
        )

        if skip_deserialization:
            api_response = api_client.ApiResponseWithoutDeserialization(response=response)
        else:
            response_for_status = _status_code_to_response.get(str(response.status))
            if response_for_status:
                api_response = response_for_status.deserialize(response, self.api_client.configuration)
            else:
                api_response = api_client.ApiResponseWithoutDeserialization(response=response)

        if not 200 <= response.status <= 299:
            raise exceptions.ApiException(
                status=response.status,
                reason=response.reason,
                api_response=api_response
            )

        return api_response


class V2DatasetsGet(BaseApi):
    # this class is used by api classes that refer to endpoints with operationId fn names

    @typing.overload
    def v2_datasets_get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def v2_datasets_get(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def v2_datasets_get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def v2_datasets_get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v2_datasets_get_oapg(
            query_params=query_params,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )


class ApiForget(BaseApi):
    # this class is used by api classes that refer to endpoints by path and http method names

    @typing.overload
    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: typing_extensions.Literal[False] = ...,
    ) -> typing.Union[
        ApiResponseFor200,
    ]: ...

    @typing.overload
    def get(
        self,
        skip_deserialization: typing_extensions.Literal[True],
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
    ) -> api_client.ApiResponseWithoutDeserialization: ...

    @typing.overload
    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = ...,
    ) -> typing.Union[
        ApiResponseFor200,
        api_client.ApiResponseWithoutDeserialization,
    ]: ...

    def get(
        self,
        query_params: RequestQueryParams = frozendict.frozendict(),
        accept_content_types: typing.Tuple[str] = _all_accept_content_types,
        stream: bool = False,
        timeout: typing.Optional[typing.Union[int, typing.Tuple]] = None,
        skip_deserialization: bool = False,
    ):
        return self._v2_datasets_get_oapg(
            query_params=query_params,
            accept_content_types=accept_content_types,
            stream=stream,
            timeout=timeout,
            skip_deserialization=skip_deserialization
        )


